#ifndef State_H
#define State_H

#include <iostream>
#include <set>
#include <list>
#include <string>
#include <deque>
#include <algorithm>

namespace hatptester{

class Entity{
private:
	const std::string type;
	const std::string name;
	
	Entity& operator=(const Entity&);
	
public:
	Entity(const std::string& type, const std::string& name):type(type), name(name){}
	
	Entity(const Entity& entity):type(entity.type), name(entity.name){}
	
	const std::string& getName() const{
		return name;
	}
	
	const std::string& getType() const{
		return type;
	}
	
	friend bool operator< (const Entity& x, const Entity& y){
		return(x.name<y.name);
	}
};

class Task{
public:
	class Parameter{
	public:
		std::string type;
		std::string name;
		std::string value;
		
		Parameter(const std::string& type, const std::string& name): type(type), name(name), value(""){}
		
		friend bool operator== (const Task::Parameter& x, const Task::Parameter& y){
			return(x.name==y.name);
		}
		
		friend bool operator== (const std::string& name, const Task::Parameter& y){
			return(name==y.name);
		}
		
		friend bool operator== (const Task::Parameter& x, const std::string& name){
			return(x.name==name);
		}
	};
	
	typedef std::vector<Parameter> Parameters; //This is to ensure that paramters are inserted in the right order
	
public:
	const std::string name;
	Parameters parameters;
	
	Task& operator=(const Task&);
	
public:
	Task(){std::cerr<<"Error : Call to an invalid constructor"<<std::endl;}
	
	Task(const std::string& name, const Parameters& parameters):name(name), parameters(parameters){}
	
	Task(const std::string& name):name(name){}
	
	Task(const Task& task):name(task.name), parameters(task.parameters){}
	
	bool addParameter(const std::string& type, const std::string& name){
		if(type=="" || name==""){
			std::cerr<<"Error : The element inserted is not vaild"<<std::endl;
			return false;
		}
		
		if(std::find(parameters.begin(), parameters.end(), name)!=parameters.end()){
			std::cerr<<"Error : The element inserted already exists"<<std::endl;
			return false;
		}
		
		parameters.push_back(Parameter(type, name));
		return true;
	}
	
	void setParameterValue(const std::string& name, const std::string& value){
		Parameters::iterator it=std::find(parameters.begin(), parameters.end(), name);
		if(it==parameters.end()){
			std::cerr<<"Error : Tries to set a parameter that doesn't exists"<<std::endl;
			return;
		}
	
		it->value=value;
	}
	
	const Parameters& getParameters() const{
		return parameters;
	}
	
	const std::string getParameterValue(const std::string& name) const{
		Parameters::const_iterator it=std::find(parameters.begin(), parameters.end(), name);
		if(it==parameters.end()){
			std::cerr<<"Error : Tries to get a parameter that doesn't exists"<<std::endl;
			return "";
		}
	
		return it->value;
	}
	
	const std::string& getName() const{
		return(name);
	}
	
	bool isReady() const{
		bool allParametersHaveValue=true;
		
		for(Parameters::const_iterator it=parameters.begin(); it!=parameters.end(); ++it){
			if(it->value==""){
				allParametersHaveValue=false;
			}
		}
		return allParametersHaveValue;
	}

	friend bool operator< (const Task& x, const Task& y){
		return(x.name<y.name);
	}
	
	friend bool operator== (const Task& x, const Task& y){
		return(x.name==y.name);
	}
};

typedef Task Method;
typedef Task Action;

class HATPState{
public:
	typedef std::map<std::string, Entity> EntitiesMap;
	typedef std::map<std::string, Task> TasksMap;
	typedef std::map<std::string, Method> MethodsMap;
	typedef std::map<std::string, Action> ActionsMap;
	
private:
	EntitiesMap entities; //Make entites unique and ordrerd by name
	MethodsMap methods; //Make methods unique and ordrerd by name
	ActionsMap actions; //Make actions unique and ordrerd by name
	
	HATPState(const HATPState&);
	HATPState& operator=(const HATPState&);

public:
	HATPState(){}
	
	HATPState(std::map<std::string, Entity> entities, std::map<std::string, Task> methods, std::map<std::string, Task> actions):entities(entities), methods(methods), actions(actions){}

	bool addEntity(const Entity& entity){
		return(entities.insert(std::make_pair(entity.getName(), entity)).second);
	}
	
	bool addMethod(const Method& method){
		return(methods.insert(std::make_pair(method.getName(), method)).second);
	}
	
	bool addAction(const Action& action){
		return(actions.insert(std::make_pair(action.getName(), action)).second);
	}

	void list() const{
		std::cout<<"### Entities"<<std::endl;
		for(EntitiesMap::const_iterator it=entities.begin(); it!=entities.end(); ++it){
			std::cout<<"\t"<<it->first<<" of type "<<it->second.getType()<<std::endl;
		}
		
		std::cout<<"### Methods"<<std::endl;
		for(MethodsMap::const_iterator it=methods.begin(); it!=methods.end(); ++it){
			std::cout<<"\t"<<it->first<<" with parameters :"<<std::endl;
			for(Task::Parameters::const_iterator it2=it->second.parameters.begin(); it2!=it->second.parameters.end(); ++it2){
				std::cout<<"\t\t\t"<<it2->name<<" of type "<<it2->type<<" = \""<<it2->value<<"\""<<std::endl;
			}
		}
		
		std::cout<<"### Actions"<<std::endl;
		for(ActionsMap::const_iterator it=actions.begin(); it!=actions.end(); ++it){
			std::cout<<"\t"<<it->first<<" with parameters :"<<std::endl;
			for(Task::Parameters::const_iterator it2=it->second.parameters.begin(); it2!=it->second.parameters.end(); ++it2){
				std::cout<<"\t\t\t"<<it2->name<<" of type "<<it2->type<<" = \""<<it2->value<<"\""<<std::endl;
			}
		}
	}

	const EntitiesMap& getEntities() const{
		return entities;
	}

	const MethodsMap& getMethods() const{
		return methods;
	}
	
	const ActionsMap& getActions() const{
		return actions;
	}
	
//TODO Try to use references instead of pointers
//	std::pair<bool, const Task&> getTask(const std::string& name){
	std::pair<bool, const Task*> getTask(const std::string& name){
//		TasksMap::const_iterator it=methods.find(name);
		TasksMap::iterator it=methods.find(name);
		if(it!=methods.end()){
//			return std::make_pair(true, it->second);
			return std::make_pair(true, &it->second);
		}
		
		it=actions.find(name);
		if(it==actions.end()){
			std::cerr<<"Error : The task "<<name<<"does not exist"<<std::endl;
//			return std::make_pair(false, Action());
			return std::pair<bool, const Task*>(false, NULL);
		}
//		return std::make_pair(true, it->second);
		return std::make_pair(true, &it->second);
	}
	
	std::pair<bool, const Method*> getMethod(const std::string& name){
		MethodsMap::iterator it=methods.find(name);
		if(it==methods.end()){
			return std::pair<bool, const Method*>(false, NULL);
		}
		return std::make_pair(true, &it->second);
	}
	
	std::pair<bool, const Action*> getAction(const std::string& name){
		ActionsMap::iterator it=actions.find(name);
		if(it==actions.end()){
			return std::pair<bool, const Action*>(false, NULL);
		}
		return std::make_pair(true, &it->second);
	}
	
	bool updateTask(const std::string& taskName, const std::string& parameterName, const std::string& value){
		TasksMap::iterator it=methods.find(taskName);
		if(it!=methods.end()){
			it->second.setParameterValue(parameterName, value);
			return true;
		}
		
		it=actions.find(taskName);
		if(it!=actions.end()){
			it->second.setParameterValue(parameterName, value);
			return true;
		}
		
		return false;
	}
	
	bool isTaskReady(const std::string name) const{
		TasksMap::const_iterator it=methods.find(name);
		if(it!=methods.end()){
			return it->second.isReady();
		}
		
		it=actions.find(name);
		if(it!=actions.end()){
			return it->second.isReady();
		}
		
		return false;
	}
};

}; //namespace

#endif
